from pyiron_lammps.decorator import calculation
from atomistics.workflows import (
    ElasticMatrixWorkflow,
    EnergyVolumeCurveWorkflow,
    optimize_positions_and_volume,
)
from atomistics.calculators import evaluate_with_lammps_library


def _optimize_structure_optional(
    lmp, structure, potential_dataframe, minimization_activated=True
):
    if minimization_activated:
        return optimize_structure(
            lmp=lmp, structure=structure, potential_dataframe=potential_dataframe
        )
    else:
        return structure


@calculation
def optimize_structure(lmp, structure, potential_dataframe):
    task_dict = optimize_positions_and_volume(structure=structure)
    structure_copy = evaluate_with_lammps_library(
        task_dict=task_dict,
        potential_dataframe=potential_dataframe,
        lmp=lmp,
        lmp_optimizer_kwargs={},
    )["structure_with_optimized_positions_and_volume"]

    # clean memory
    lmp.interactive_lib_command("clear")
    return structure_copy


@calculation
def calculate_elastic_constants(
    lmp,
    structure,
    potential_dataframe,
    num_of_point=5,
    eps_range=0.005,
    sqrt_eta=True,
    fit_order=2,
    minimization_activated=False,
):
    # Optimize structure
    structure_opt = _optimize_structure_optional(
        lmp=lmp,
        structure=structure,
        potential_dataframe=potential_dataframe,
        minimization_activated=minimization_activated,
    )

    # Generate structures
    calculator = ElasticMatrixWorkflow(
        structure=structure_opt.copy(),
        num_of_point=num_of_point,
        eps_range=eps_range,
        sqrt_eta=sqrt_eta,
        fit_order=fit_order,
    )
    structure_dict = calculator.generate_structures()

    # run calculation
    energy_tot_lst = evaluate_with_lammps_library(
        task_dict=structure_dict,
        potential_dataframe=potential_dataframe,
        lmp=lmp,
        lmp_optimizer_kwargs={},
    )

    # fit
    result_dict = calculator.analyse_structures(
        output_dict=energy_tot_lst,
        output_keys=("elastic_matrix",),
    )
    return result_dict["elastic_matrix"]


@calculation
def calculate_energy_volume_curve(
    lmp,
    structure,
    potential_dataframe,
    num_points=11,
    fit_type="polynomial",
    fit_order=3,
    vol_range=0.05,
    axes=("x", "y", "z"),
    strains=None,
    minimization_activated=False,
):
    # Optimize structure
    structure_opt = _optimize_structure_optional(
        lmp=lmp,
        structure=structure,
        potential_dataframe=potential_dataframe,
        minimization_activated=minimization_activated,
    )

    # Generate structures
    calculator = EnergyVolumeCurveWorkflow(
        structure=structure_opt.copy(),
        num_points=num_points,
        fit_type=fit_type,
        fit_order=fit_order,
        vol_range=vol_range,
        axes=axes,
        strains=strains,
    )
    structure_dict = calculator.generate_structures()

    # run calculation
    energy_tot_lst = evaluate_with_lammps_library(
        task_dict=structure_dict,
        potential_dataframe=potential_dataframe,
        lmp=lmp,
    )

    # fit
    calculator.analyse_structures(energy_tot_lst)
    return calculator.fit_dict



import os
import unittest
from ase.build import bulk
import pyiron_lammps as pyr
import structuretoolkit as stk


def validate_elastic_constants(elastic_matrix):
    return [
        elastic_matrix[0, 0] > 200,
        elastic_matrix[1, 1] > 200,
        elastic_matrix[2, 2] > 200,
        elastic_matrix[0, 1] > 135,
        elastic_matrix[0, 2] > 135,
        elastic_matrix[1, 2] > 135,
        elastic_matrix[3, 3] > 100,
        elastic_matrix[4, 4] > 100,
        elastic_matrix[5, 5] > 100,
    ]


class TestParallelSingleCore(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        count_lst = [22, 22, 22, 21, 21]
        element_lst = ["Fe", "Ni", "Cr", "Co", "Cu"]
        potential = '2021--Deluigi-O-R--Fe-Ni-Cr-Co-Cu--LAMMPS--ipr1'
        resource_path = os.path.join(os.path.dirname(__file__), "static")

        # Generate SQS Structure
        structure_template = bulk("Al", cubic=True).repeat([3, 3, 3])
        mole_fractions = {
            el: c / len(structure_template) for el, c in zip(element_lst, count_lst)
        }
        structure = stk.build.sqs_structures(
            structure=structure_template,
            mole_fractions=mole_fractions,
        )[0]

        # Select potential
        df_pot = pyr.get_potential_dataframe(
            structure=structure,
            resource_path=resource_path
        )

        # Assign variable
        cls.df_pot_selected = df_pot[df_pot.Name == potential].iloc[0]
        cls.structure = structure
        cls.resource_path = resource_path
        cls.potential = potential
        cls.count_lst = count_lst

    def test_structure(self):
        self.assertEqual(len(self.structure), sum(self.count_lst))

    def test_example_elastic_constants_parallel_cores_one(self):
        structure_opt_lst = pyr.optimize_structure(
            structure=[self.structure.copy()],
            potential_dataframe=self.df_pot_selected,
            executor=None,
        )

        # Calculate Elastic Constants
        elastic_matrix = pyr.calculate_elastic_constants(
            structure=structure_opt_lst,
            potential_dataframe=self.df_pot_selected,
            num_of_point=5,
            eps_range=0.005,
            sqrt_eta=True,
            fit_order=2,
            executor=None,
            minimization_activated=False,
        )[0]

        self.assertEqual(len(structure_opt_lst[0]), sum(self.count_lst))
        self.assertTrue(all(validate_elastic_constants(elastic_matrix=elastic_matrix)))

    def test_example_elastic_constants_with_minimization_parallel_cores_one(self):
        elastic_matrix = pyr.calculate_elastic_constants(
            structure=[self.structure.copy()],
            potential_dataframe=self.df_pot_selected,
            num_of_point=5,
            eps_range=0.005,
            sqrt_eta=True,
            fit_order=2,
            executor=None,
            minimization_activated=True,
        )[0]
        self.assertTrue(all(validate_elastic_constants(elastic_matrix=elastic_matrix)))


if __name__ == '__main__':
    unittest.main()


import os
import unittest
from ase.build import bulk
import pyiron_lammps as pyr
import structuretoolkit as stk
from pympipool import Executor


def validate_fitdict(fit_dict):
    lst = [
        fit_dict['b_prime_eq'] > 1.5,
        fit_dict['b_prime_eq'] < 3.0,
        fit_dict['bulkmodul_eq'] > 174,
        fit_dict['bulkmodul_eq'] < 184,
        fit_dict['energy_eq'] > -453.9,
        fit_dict['energy_eq'] < -453.5,
        fit_dict['volume_eq'] > 1207,
        fit_dict['volume_eq'] < 1213,
    ]
    if not all(lst):
        print(fit_dict)
    return lst


class TestParallelTwoCores(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        count_lst = [22, 22, 22, 21, 21]
        element_lst = ["Fe", "Ni", "Cr", "Co", "Cu"]
        potential = '2021--Deluigi-O-R--Fe-Ni-Cr-Co-Cu--LAMMPS--ipr1'
        resource_path = os.path.join(os.path.dirname(__file__), "static")

        # Generate SQS Structure
        structure_template = bulk("Al", cubic=True).repeat([3, 3, 3])
        mole_fractions = {
            el: c / len(structure_template) for el, c in zip(element_lst, count_lst)
        }
        structure = stk.build.sqs_structures(
            structure=structure_template,
            mole_fractions=mole_fractions,
        )[0]

        # Select potential
        df_pot = pyr.get_potential_dataframe(
            structure=structure,
            resource_path=resource_path
        )

        # Assign variable
        cls.df_pot_selected = df_pot[df_pot.Name == potential].iloc[0]
        cls.structure = structure
        cls.resource_path = resource_path
        cls.potential = potential
        cls.count_lst = count_lst

    def test_structure(self):
        self.assertEqual(len(self.structure), sum(self.count_lst))

    def test_example_elastic_constants_parallel_cores_two(self):
        with Executor(max_workers=2, hostname_localhost=True) as exe:
            structure_opt_lst = pyr.optimize_structure(
                structure=[self.structure.copy()],
                potential_dataframe=[self.df_pot_selected],
                executor=exe,
            )

            # Calculate Elastic Constants
            fit_dict = pyr.calculate_energy_volume_curve(
                structure=structure_opt_lst,
                potential_dataframe=[self.df_pot_selected],
                num_points=11,
                fit_type="polynomial",
                fit_order=3,
                vol_range=0.05,
                axes=["x", "y", "z"],
                strains=None,
                executor=exe,
                minimization_activated=False,
            )[0]

        self.assertEqual(len(structure_opt_lst[0]), sum(self.count_lst))
        self.assertTrue(all(validate_fitdict(fit_dict=fit_dict)))

    def test_example_elastic_constants_with_minimization_parallel_cores_two(self):
        with Executor(max_workers=2, hostname_localhost=True) as exe:
            fit_dict = pyr.calculate_energy_volume_curve(
                structure=[self.structure.copy()],
                potential_dataframe=[self.df_pot_selected],
                num_points=11,
                fit_type="polynomial",
                fit_order=3,
                vol_range=0.05,
                axes=["x", "y", "z"],
                strains=None,
                executor=exe,
                minimization_activated=True,
            )[0]
        self.assertTrue(all(validate_fitdict(fit_dict=fit_dict)))


if __name__ == '__main__':
    unittest.main()


import numpy as np
from ase.atoms import Atoms
from pandas import DataFrame, Series
from pylammpsmpi import LammpsASELibrary

from pyiron_lammps.calculation import (
    optimize_structure,
    calculate_elastic_constants,
    calculate_energy_volume_curve,
)


def _get_lammps_mpi(enable_mpi=True):
    if enable_mpi:
        # To get the right instance of MPI.COMM_SELF it is necessary to import it inside the function.
        from mpi4py import MPI

        return LammpsASELibrary(
            working_directory=None,
            cores=1,
            comm=MPI.COMM_SELF,
            logger=None,
            log_file=None,
            library=None,
            diable_log_file=True,
        )
    else:
        return LammpsASELibrary(
            working_directory=None,
            cores=1,
            comm=None,
            logger=None,
            log_file=None,
            library=None,
            diable_log_file=True,
        )


def _parallel_execution(function, input_parameter_lst, lmp=None, executor=None):
    if executor is None:
        return [
            function(input_parameter=input_parameter + [False, lmp])
            for input_parameter in input_parameter_lst
        ]
    elif executor is not None and lmp is None:
        return list(
            executor.map(
                function,
                [
                    input_parameter + [True, None]
                    for input_parameter in input_parameter_lst
                ],
            )
        )
    elif executor is not None and lmp is not None:
        raise ValueError(
            "The external LAMMPS instance can only be used for serial execution."
        )
    else:
        raise ValueError("The number of cores has to be a positive integer.")


def _optimize_structure_serial(input_parameter):
    structure, potential_dataframe, enable_mpi, lmp = input_parameter
    if lmp is None:
        return optimize_structure(
            lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
            structure=structure,
            potential_dataframe=potential_dataframe,
        )
    else:
        return optimize_structure(
            lmp=lmp,
            structure=structure,
            potential_dataframe=potential_dataframe,
        )


def _calculate_elastic_constants_serial(input_parameter):
    (
        structure,
        potential_dataframe,
        num_of_point,
        eps_range,
        sqrt_eta,
        fit_order,
        minimization_activated,
        enable_mpi,
        lmp,
    ) = input_parameter
    if lmp is None:
        return calculate_elastic_constants(
            lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
            structure=structure,
            potential_dataframe=potential_dataframe,
            num_of_point=num_of_point,
            eps_range=eps_range,
            sqrt_eta=sqrt_eta,
            fit_order=fit_order,
            minimization_activated=minimization_activated,
        )
    else:
        return calculate_elastic_constants(
            lmp=lmp,
            structure=structure,
            potential_dataframe=potential_dataframe,
            num_of_point=num_of_point,
            eps_range=eps_range,
            sqrt_eta=sqrt_eta,
            fit_order=fit_order,
            minimization_activated=minimization_activated,
        )


def _calculate_energy_volume_curve_serial(input_parameter):
    (
        structure,
        potential_dataframe,
        num_points,
        fit_type,
        fit_order,
        vol_range,
        axes,
        strains,
        minimization_activated,
        enable_mpi,
        lmp,
    ) = input_parameter
    if lmp is None:
        return calculate_energy_volume_curve(
            lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
            structure=structure,
            potential_dataframe=potential_dataframe,
            num_points=num_points,
            fit_type=fit_type,
            fit_order=fit_order,
            vol_range=vol_range,
            axes=axes,
            strains=strains,
            minimization_activated=minimization_activated,
        )
    else:
        return calculate_energy_volume_curve(
            lmp=_get_lammps_mpi(enable_mpi=enable_mpi),
            structure=structure,
            potential_dataframe=potential_dataframe,
            num_points=num_points,
            fit_type=fit_type,
            fit_order=fit_order,
            vol_range=vol_range,
            axes=axes,
            strains=strains,
            minimization_activated=minimization_activated,
        )


def optimize_structure_parallel(
    structure, potential_dataframe, lmp=None, executor=None
):
    input_parameter_lst, output_as_lst = combine_structure_and_potential(
        structure=structure, potential_dataframe=potential_dataframe
    )
    if output_as_lst:
        return _parallel_execution(
            function=_optimize_structure_serial,
            input_parameter_lst=input_parameter_lst,
            lmp=lmp,
            executor=executor,
        )
    else:
        return _parallel_execution(
            function=_optimize_structure_serial,
            input_parameter_lst=input_parameter_lst,
            lmp=lmp,
            executor=executor,
        )[0]


def calculate_elastic_constants_parallel(
    structure,
    potential_dataframe,
    num_of_point=5,
    eps_range=0.005,
    sqrt_eta=True,
    fit_order=2,
    minimization_activated=False,
    executor=None,
    lmp=None,
):
    combo_lst, output_as_lst = combine_structure_and_potential(
        structure=structure, potential_dataframe=potential_dataframe
    )
    input_parameter_lst = [
        [
            s,
            p,
            num_of_point,
            eps_range,
            sqrt_eta,
            fit_order,
            minimization_activated,
        ]
        for s, p in combo_lst
    ]
    if output_as_lst:
        return _parallel_execution(
            function=_calculate_elastic_constants_serial,
            input_parameter_lst=input_parameter_lst,
            executor=executor,
            lmp=lmp,
        )
    else:
        return _parallel_execution(
            function=_calculate_elastic_constants_serial,
            input_parameter_lst=input_parameter_lst,
            executor=executor,
            lmp=lmp,
        )[0]


def calculate_energy_volume_curve_parallel(
    structure,
    potential_dataframe,
    num_points=11,
    fit_type="polynomial",
    fit_order=3,
    vol_range=0.05,
    axes=("x", "y", "z"),
    strains=None,
    minimization_activated=False,
    executor=None,
    lmp=None,
):
    combo_lst, output_as_lst = combine_structure_and_potential(
        structure=structure, potential_dataframe=potential_dataframe
    )
    input_parameter_lst = [
        [
            s,
            p,
            num_points,
            fit_type,
            fit_order,
            vol_range,
            axes,
            strains,
            minimization_activated,
        ]
        for s, p in combo_lst
    ]
    if output_as_lst:
        return _parallel_execution(
            function=_calculate_energy_volume_curve_serial,
            input_parameter_lst=input_parameter_lst,
            executor=executor,
            lmp=lmp,
        )
    else:
        return _parallel_execution(
            function=_calculate_energy_volume_curve_serial,
            input_parameter_lst=input_parameter_lst,
            executor=executor,
            lmp=lmp,
        )[0]


def combine_structure_and_potential(structure, potential_dataframe):
    if isinstance(structure, (list, np.ndarray)):
        if isinstance(potential_dataframe, (list, np.ndarray)):
            if len(structure) == len(potential_dataframe):
                return [[s, p] for s, p in zip(structure, potential_dataframe)], True
            else:
                raise ValueError(
                    "Input lists have len(structure) != len(potential_dataframe) ."
                )
        elif isinstance(potential_dataframe, (DataFrame, Series)):
            return [[s, potential_dataframe] for s in structure], True
        else:
            raise TypeError(
                "potential_dataframe should either be an pandas.DataFrame object or a list of those. "
            )
    elif isinstance(structure, Atoms):
        if isinstance(potential_dataframe, (list, np.ndarray)):
            return [[structure, p] for p in potential_dataframe], True
        elif isinstance(potential_dataframe, (DataFrame, Series)):
            return [[structure, potential_dataframe]], False
        else:
            raise TypeError(
                "potential_dataframe should either be an pandas.DataFrame object or a list of those. "
            )
    else:
        raise TypeError(
            "structure should either be an ase.atoms.Atoms object or a list of those."
        )
